# Sample run of importing the code:
# -------------------------------------
#
#             from PAMI.fuzzyPeriodicFrequentPattern.basic import FPFPMiner as alg
#
#             obj =alg.FPFPMiner("input.txt",2,3)
#
#             obj.mine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Fuzzy Periodic Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.save("output.txt")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from PAMI.fuzzyPeriodicFrequentPattern.basic import abstract as _ab
from deprecated import deprecated


class _FFList:
    """
    A class represent a Fuzzy List of an element

    :Attributes:

        item : int
            the item name
        sumLUtil : float
            the sum of utilities of a fuzzy item in database
        sumRUtil : float
            the sum of resting values of a fuzzy item in database
        elements : list
            list of elements contain tid,Utility and resting values of element in each transaction
        maxPeriod : int
            it represents the max period of a item

    :Methods:

        addElement(element)
            Method to add an element to this fuzzy list and update the sums at the same time.
        printElement(e)
            Method to print elements

    """

    def __init__(self, itemName):
        self.item = itemName
        self.sumLUtil = 0.0
        self.sumRUtil = 0.0
        self.elements = []
        self.maxPeriod = 0

    def addElement(self, element):
        """
        A Method that add a new element to FFList

        :param element: an element to be added to FFList
        :type element: Element
        """
        self.sumLUtil += element.lUtils
        self.sumRUtil += element.rUtils
        self.elements.append(element)
        self.maxPeriod = max(self.maxPeriod, element.period)

    def printElement(self):
        """
        A Method to Print elements in the FFList
        """
        for ele in self.elements:
            print(ele.tid, ele.lUtils, ele.rUtils, ele.period)


class _Element:
    """
    A class represents an Element of a fuzzy list

    :Attributes:

        tid : int
            keep tact of transaction id
        lUtils : float
            the utility of a fuzzy item in the transaction
        rUtils : float
            the resting value of a fuzzy item in the transaction
        period : int
            represent the period of the element
    """

    def __init__(self, tid, iUtil, rUtil, period):
        self.tid = tid
        self.lUtils = iUtil
        self.rUtils = rUtil
        self.period = period


class _Regions:
    """
    A class calculate the regions

    :Attributes:

        low : int
            low region value
        middle : int
            middle region value
        high : int
            high region values
        """

    def __init__(self, quantity, regionsNumber):
        self.low = 0
        self.middle = 0
        self.high = 0
        if regionsNumber == 3:  # if we have 3 regions
            if 0 < quantity <= 1:
                self.low = 1
                self.high = 0
                self.middle = 0
            elif 1 < quantity <= 6:
                self.low = float((6 - quantity) / 5)
                self.middle = float((quantity - 1) / 5)
                self.high = 0
            elif 6 < quantity <= 11:
                self.low = 0
                self.middle = float((11 - quantity) / 5)
                self.high = float((quantity - 6) / 5)
            else:
                self.low = 0
                self.middle = 0
                self.high = 1


class _Pair:
    """
    A class to store item name and quantity together.
    """

    def __init__(self):
        self.item = 0
        self.quantity = 0


class FPFPMiner(_ab._fuzzyPeriodicFrequentPatterns):
    """
    :Description:   Fuzzy Periodic Frequent Pattern Miner is desired to find all fuzzy periodic frequent patterns which is
                    on-trivial and challenging problem to its huge search space.we are using efficient pruning
                    techniques to reduce the search space.

    :Reference:


    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent patterns
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minSup: int or float or str :
                   The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.
    :param maxPer: float :
                   The user can specify maxPer in count or proportion of database size. If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

        iFile : file
            Name of the input file to mine complete set of fuzzy spatial frequent patterns
        oFile : file
               Name of the oFile file to store complete set of fuzzy spatial frequent patterns
        minSup : float
            The user given support
        period : int
            periodicity of an element
        memoryRSS : float
                To store the total amount of RSS memory consumed by the program
        startTime : float
               To record the start time of the mining process
        endTime : float
            To record the completion time of the mining process
        itemsCnt : int
            To record the number of fuzzy spatial itemSets generated
        mapItemsLowSum : map
            To keep track of low region values of items
        mapItemsMidSum : map
            To keep track of middle region values of items
        mapItemsHighSum : map
            To keep track of high region values of items
        mapItemSum : map
            To keep track of sum of Fuzzy Values of items
        mapItemRegions : map
            To Keep track of fuzzy regions of item
        joinsCnt : int
            To keep track of the number of FFI-list that was constructed
        BufferSize : int
            represent the size of Buffer
        itemSetBuffer list
            to keep track of items in buffer
        maxTID : int
            represent the maximum tid of the database
        lastTIDs : map
            represent the last tid of fuzzy items
        itemsToRegion : map
            represent items with respective regions

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        convert(value)
            To convert the given user specified value
        FSFIMining( prefix, prefixLen, fsFim, minSup)
            Method generate FFI from prefix
        construct(px, py)
            A function to construct Fuzzy itemSet from 2 fuzzy itemSets
        findElementWithTID(UList, tid)
            To find element with same tid as given
        WriteOut(prefix, prefixLen, item, sumIUtil,period)
            To Store the patten

    **Executing the code on terminal :**
    ---------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 FPFPMiner_old.py <inputFile> <outputFile> <minSup> <maxPer> <sep>

      Example Usage:

      (.venv) $ python3  FPFPMiner_old.py sampleTDB.txt output.txt 2 3

    .. note:: minSup will be considered in percentage of database transactions


    **Sample run of importing the code:**
    --------------------------------------

        from PAMI.fuzzyPeriodicFrequentPattern.basic import FPFPMiner as alg

        obj =alg.FPFPMiner("input.txt",2,3)

        obj.mine()

        periodicFrequentPatterns = obj.getPatterns()

        print("Total number of Fuzzy Periodic Frequent Patterns:", len(periodicFrequentPatterns))

        obj.save("output.txt")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ---------------
            The complete program was written by Sai Chitra.B under the supervision of Professor Rage Uday Kiran.

    """
    _startTime = float()
    _endTime = float()
    _minSup = str()
    _maxPer = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _sep = " "
    _Database = []
    _transactions = []
    _fuzzyValues = []
    _ts = []

    def __init__(self, iFile, minSup, period, sep="\t"):
        super().__init__(iFile, minSup, period, sep)
        self._oFile = ""
        self._BufferSize = 200
        self._itemSetBuffer = []
        self._mapItemRegions = {}
        self._mapItemSum = {}
        self._mapItemsHighSum = {}
        self._finalPatterns = {}
        self._joinsCnt = 0
        self._itemsCnt = 0
        self._mapItemMidSum = {}
        self._startTime = float()
        self._endTime = float()
        self._mapItemsLowSum = {}
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._dbLen = 0

    def _compareItems(self, o1, o2):
        """
        A Function that sort all FFI-list in ascending order of Support

        :param o1: First FFI-list
        :type o1: _FFList
        :param o2: Second FFI-list
        :type o2: _FFList
        :return: Comparison Value
        :rtype: int
        """
        compare = self._mapItemSum[o1.item] - self._mapItemSum[o2.item]
        if compare == 0:
            return int(o1.item) - int(o2.item)
        else:
            return compare

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :type value: int  or float or str
        :return: converted value
        :rtype: float
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbLen * value)
        if type(value) is str:
            if '.' in value:
                value = (self._dbLen * value)
            else:
                value = int(value)
        return value

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        data, self._transactions, self._fuzzyValues, ts = [], [], [], []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                self._ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                self._transactions = self._iFile['Transactions'].tolist()
            if 'fuzzyValues' in i:
                self._fuzzyValues = self._iFile['fuzzyValues'].tolist()
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                count = 0
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = line.split(":")
                    parts[0] = parts[0].strip()
                    parts[2] = parts[2].strip()
                    items = parts[0].split(self._sep)
                    quantities = parts[2].split(self._sep)
                    self._ts.append(count)
                    self._transactions.append([x for x in items])
                    self._fuzzyValues.append([x for x in quantities])
                    count += 1
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        count = 0
                        for line in f:
                            line = line.split("\n")[0]
                            parts = line.split(":")
                            parts[0] = parts[0].strip()
                            parts[2] = parts[2].strip()
                            items = parts[0].split(self._sep)
                            quantities = parts[2].split(self._sep)
                            self._ts.append(count)
                            self._transactions.append([x for x in items])
                            self._fuzzyValues.append([x for x in quantities])
                            count += 1
                except IOError:
                    print("File Not Found")
                    quit()

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Fuzzy periodic Frequent pattern mining process will start from here
        """
        self.mine()

    def mine(self):
        """
        Fuzzy periodic Frequent pattern mining process will start from here
        """
        maxTID = 0
        lastTIDs = {}
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._finalPatterns = {}
        tid = int()
        for line in range(len(self._transactions)):
            tid = int(self._ts[line])
            self._dbLen += 1
            items = self._transactions[line]
            quantities = self._fuzzyValues[line]
            if tid < maxTID:
                maxTID = tid
            for i in range(0, len(items)):
                regions = _Regions(int(quantities[i]), 3)
                item = items[i]
                if item in self._mapItemsLowSum.keys():
                    low = self._mapItemsLowSum[item]
                    low += regions.low
                    self._mapItemsLowSum[item] = low
                else:
                    self._mapItemsLowSum[item] = regions.low
                if item in self._mapItemMidSum.keys():
                    mid = self._mapItemMidSum[item]
                    mid += regions.middle
                    self._mapItemMidSum[item] = mid
                else:
                    self._mapItemMidSum[item] = regions.middle
                if item in self._mapItemsHighSum.keys():
                    high = self._mapItemsHighSum[item]
                    high += regions.high
                    self._mapItemsHighSum[item] = high
                else:
                    self._mapItemsHighSum[item] = regions.high
        listOfFFIList = []
        mapItemsToFFLIST = {}
        itemsToRegion = {}
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        for item1 in self._mapItemsLowSum.keys():
            item = item1
            low = self._mapItemsLowSum[item]
            mid = self._mapItemMidSum[item]
            high = self._mapItemsHighSum[item]
            if low >= mid and low >= high:
                self._mapItemSum[item] = low
                self._mapItemRegions[item] = "L"
                itemsToRegion[item] = "L"
            elif mid >= low and mid >= high:
                self._mapItemSum[item] = mid
                self._mapItemRegions[item] = "M"
                itemsToRegion[item] = "M"
            elif high >= low and high >= mid:
                self._mapItemRegions[item] = "H"
                self._mapItemSum[item] = high
                itemsToRegion[item] = "H"
            if self._mapItemSum[item] >= self._minSup:
                fUList = _FFList(item)
                k = tuple([item, itemsToRegion.get(item)])
                mapItemsToFFLIST[k] = fUList
                listOfFFIList.append(fUList)
                lastTIDs[item] = tid
        listOfFFIList.sort(key=_ab._functools.cmp_to_key(self._compareItems))
        for line in range(len(self._transactions)):
            tid = int(self._ts[line])
            items = self._transactions[line]
            quantities = self._fuzzyValues[line]
            revisedTransaction = []
            for i in range(0, len(items)):
                pair = _Pair()
                pair.item = items[i]
                regions = _Regions(int(quantities[i]), 3)
                item = pair.item
                if self._mapItemSum[item] >= self._minSup:
                    if self._mapItemRegions[pair.item] == "L":
                        pair.quantity = regions.low
                    elif self._mapItemRegions[pair.item] == "M":
                        pair.quantity = regions.middle
                    elif self._mapItemRegions[pair.item] == "H":
                        pair.quantity = regions.high
                    if pair.quantity > 0:
                        revisedTransaction.append(pair)
            revisedTransaction.sort(key=_ab._functools.cmp_to_key(self._compareItems))
            for i in range(len(revisedTransaction) - 1, -1, -1):
                pair = revisedTransaction[i]
                remainUtil = 0
                for j in range(len(revisedTransaction) - 1, i - 1, -1):
                    remainUtil += revisedTransaction[j].quantity
                if pair.quantity > remainUtil:
                    remainingUtility = pair.quantity
                else:
                    remainingUtility = remainUtil
                if mapItemsToFFLIST.get(tuple([pair.item, itemsToRegion[pair.item]])) is not None:
                    FFListOfItem = mapItemsToFFLIST[tuple([pair.item, itemsToRegion[pair.item]])]
                    if len(FFListOfItem.elements) == 0:
                        element = _Element(tid, pair.quantity, remainingUtility, 0)
                    else:
                        if lastTIDs[pair.item] == tid:
                            element = _Element(tid, pair.quantity, remainingUtility, maxTID - tid)
                        else:
                            lastTid = FFListOfItem.elements[-1].tid
                            curPer = tid - lastTid
                            element = _Element(tid, pair.quantity, remainingUtility, curPer)
                    FFListOfItem.addElement(element)
        self._FSFIMining(self._itemSetBuffer, 0, listOfFFIList, self._minSup)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def _FSFIMining(self, prefix, prefixLen, fsFim, minSup):

        """
        Generates FPFP from prefix

        :param prefix: the prefix patterns of FPFP
        :type prefix: len
        :param prefixLen: the length of prefix
        :type prefixLen: int
        :param fsFim: the Fuzzy list of prefix itemSets
        :type fsFim: list
        :param minSup: the minimum support of
        :type minSup:int
        """
        for i in range(0, len(fsFim)):
            X = fsFim[i]
            if X.sumLUtil >= minSup and X.maxPeriod <= self._maxPer:
                self._WriteOut(prefix, prefixLen, X.item, X.sumLUtil, X.maxPeriod)
            if X.sumRUtil >= minSup:
                exULs = []
                for j in range(i + 1, len(fsFim)):
                    Y = fsFim[j]
                    exULs.append(self._construct(X, Y))
                    self._joinsCnt += 1
                self._itemSetBuffer.insert(prefixLen, X.item)
                self._FSFIMining(self._itemSetBuffer, prefixLen + 1, exULs, minSup, )

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime - self._startTime

    def _construct(self, px, py):
        """
        A function to construct a new Fuzzy item set from 2 fuzzy itemSets

        :param px: the item set px
        :type px: FFI-List
        :param py: item set py
        :type py: FFI-List
        :return: the item set of pxy(px and py)
        :rtype: FFI-List
        """
        pxyUL = _FFList(py.item)
        prev = 0
        for ex in px.elements:
            ey = self._findElementWithTID(py, ex.tid)
            if ey is None:
                continue
            eXY = _Element(ex.tid, min([ex.lUtils, ey.lUtils], key=lambda x: float(x)), ey.rUtils, ex.tid - prev)
            pxyUL.addElement(eXY)
            prev = ex.tid
        return pxyUL

    def _findElementWithTID(self, UList, tid):
        """
        To find element with same tid as given

        :param UList: fuzzy list
        :type UList: FFI-List
        :param tid: transaction id
        :type tid: int
        :return: element with tid as given
        :rtype: element if exist or None
        """
        List = UList.elements
        first = 0
        last = len(List) - 1
        while first <= last:
            mid = (first + last) >> 1
            if List[mid].tid < tid:
                first = mid + 1
            elif List[mid].tid > tid:
                last = mid - 1
            else:
                return List[mid]
        return None

    def _WriteOut(self, prefix, prefixLen, item, sumLUtil, period):
        """
        To Store the patten

        :param prefix: prefix of itemSet
        :type prefix: list
        :param prefixLen: length of prefix
        :type prefixLen: int
        :param item: the last item
        :type item: int
        :param sumLUtil: sum of utility of itemSet
        :type sumLUtil: float
        :param period: represent the period of itemSet
        :type period: int
        """
        self._itemsCnt += 1
        res = ""
        for i in range(0, prefixLen):
            res += str(prefix[i]) + "." + str(self._mapItemRegions[prefix[i]]) + "\t"
        res += str(item) + "." + str(self._mapItemRegions.get(item))
        #res1 = str(sumLUtil) + " : " + str(period)
        self._finalPatterns[res] = [sumLUtil, period]

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv ile
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.strip() + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % patternsAndSupport)

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Fuzzy Periodic-Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:  # to  include a user specified separator
            _ap = FPFPMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:  # to consider "\t" as a separator
            _ap = FPFPMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.mine()
        _ap.mine()
        print("Total number of Fuzzy Periodic-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")




# VBFTMine is one of the fundamental algorithm to discover fault-tolerant frequent patterns in an uncertain transactional database based on bitset representation.
#
# **Importing this algorithm into a python program**
#
#             import PAMI.uncertainFaultTolerantFrequentPattern.basic.VBFTMine as alg
#
#             iFile = 'sampleDB.txt'
#
#             minSup = 10  # can also be specified between 0 and 1
#
#             itemSup = 2  # can also be specified between 0 and 1
#
#             minLength = 3 # can also be specified between 0 and 1
#
#             faultTolerance = 2 # can also be specified between 0 and 1
#
#             obj = alg.VBFTMine(iFile, minSup, itemSup, minLength, faultTolerance)
#
#             obj.mine()
#
#             faultTolerantFrequentPattern = obj.getPatterns()
#
#             print("Total number of Fault Tolerant Frequent Patterns:", len(faultTolerantFrequentPattern))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#

__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
from deprecated import deprecated

import numpy as _np
from PAMI.faultTolerantFrequentPattern.basic import abstract as _ab

class VBFTMine(_ab._faultTolerantFrequentPatterns):
    """
    About this algorithm
    ====================
    
    :Description:  VBFTMine is one of the fundamental algorithm to discover fault tolerant frequent patterns in an uncertain transactional database based on
                   bitset representation.
                   This program employs apriori property (or downward closure property) to  reduce the search space effectively.

    :Reference:   Koh, JL., Yo, PW. (2005). An Efficient Approach for Mining Fault-Tolerant Frequent Patterns Based on Bit Vector Representations.
            In:   Zhou, L., Ooi, B.C., Meng, X. (eds) Database Systems for Advanced Applications. DASFAA 2005. Lecture Notes in Computer Science,
                  vol 3453. Springer, Berlin, Heidelberg. https://doi.org/10.1007/11408079_51

    :param  iFile: str :
                   Name of the Input file to mine complete set of uncertain Fault Tolerant FrequentFrequent Patterns
    :param  oFile: str :
                   Name of the output file to store complete set of uncertain Fault Tolerant FrequentFrequent Patterns
    :param  minSup: float or int or str :
                   The user can specify minSup either in count or proportion of database size.
                   If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                   Otherwise, it will be treated as float.
                   Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
    :param  itemSup: int or float :
                    Frequency of an item
    :param minLength: int
                    minimum length of a pattern
    :param faultTolerance: int :
                    The ability of a pattern mining algorithm to handle errors or inconsistencies in the data without completely failing or producing incorrect results.
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        startTime : float
          To record the start time of the mining process

        endTime : float
          To record the completion time of the mining process

        finalPatterns : dict
          Storing the complete set of patterns in a dictionary variable

        memoryUSS : float
          To store the total amount of USS memory consumed by the program

        memoryRSS : float
          To store the total amount of RSS memory consumed by the program

        Database : list
          To store the transactions of a database in list


    Execution methods
    =================

    **Terminal command**


    .. code-block:: console

       Format:

       (.venv) $ python3 VBFTMine.py <inputFile> <outputFile> <minSup> <itemSup> <minLength> <faultTolerance>

       Examples usage:

       (.venv) $ python3 VBFTMine.py sampleDB.txt patterns.txt 10.0 3.0 3 1

    .. note:: minSup will be considered in times of minSup and count of database transactions


    **Calling from a python program**

    .. code-block:: python
    
            import PAMI.faultTolerantFrequentPattern.basic.VBFTMine as alg

            iFile = 'sampleDB.txt'

            minSup = 10  # can also be specified between 0 and 1

            itemSup = 2  # can also be specified between 0 and 1

            minLength = 3 # can also be specified between 0 and 1

            faultTolerance = 2 # can also be specified between 0 and 1

            obj = alg.VBFTMine(iFile, minSup, itemSup, minLength, faultTolerance)

            obj.mine()

            faultTolerantFrequentPattern = obj.getPatterns()

            print("Total number of Fault Tolerant Frequent Patterns:", len(faultTolerantFrequentPattern))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    Credits
    =======

           The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

    """

    _minSup = float()
    _itemSup = float()
    _minLength = int()
    _faultTolerance = int()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _plist = []
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _mapSupport = {}

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()

            for k in temp:
                self._Database.append(set(k))
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(set(temp))
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            for i in temp:
                                if i not in self._plist:
                                    self._plist.append(i)
                            self._Database.append(set(temp))
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _Count(self, tids):
        """
        Count the occurrences of 1s in the given list of transaction IDs.

        :param tids: List of transaction IDs.
        :type tids: List[int]
        :return: Count of occurrences of 1s in the list.
        :rtype: int
        """
        count = 0
        for i in tids:
            if i == 1:
                count += 1
        return count

    def _save(self, prefix, suffix, tidsetx):
        """
        Save the pattern with its support count if it meets the fault tolerance criteria.

        :param prefix: Prefix part of the pattern.
        :type prefix: list
        :param suffix: Suffix part of the pattern.
        :type suffix: list
        :param tidsetx: Transaction IDs associated with the pattern.
        :type tidsetx: list
        :return: None
        """
        if (prefix == None):
            prefix = suffix
        else:
            prefix = prefix + suffix
        prefix = list(set(prefix))
        prefix.sort()
        val = self._Count(tidsetx)
        if len(prefix) > self._faultTolerance:
            self._finalPatterns[tuple(prefix)] = val

    def _processEquivalenceClass(self, prefix, itemsets, tidsets):
        """
        Process the equivalence class to generate frequent patterns.

        :param prefix: Prefix part of the pattern.
        :type prefix: list.
        :param itemsets: List of itemsets in the equivalence class.
        :type itemsets: list.
        :param tidsets: List of transaction IDs associated with each itemset.
        :type tidsets: list
        :return: None
        """
        if (len(itemsets) == 1):
            i = itemsets[0]
            tidi = tidsets[0]
            self._save(prefix, [i], tidi)
            return
        for i in range(len(itemsets)):
            itemx = itemsets[i]
            if (itemx == None):
                continue
            tidsetx = tidsets[i]
            classItemsets = []
            classtidsets = []
            itemsetx = [itemx]
            for j in range(i + 1, len(itemsets)):
                itemj = itemsets[j]
                tidsetj = tidsets[j]
                y = list(_np.array(tidsetx) & _np.array(tidsetj))
                total = self._Count(y)
                if total >= self._minSup:
                    classItemsets.append(itemj)
                    classtidsets.append(y)
            if (len(classItemsets) > 0):
                newprefix = list(set(itemsetx)) + prefix
                self._processEquivalenceClass(newprefix, classItemsets, classtidsets)
            self._save(prefix, list(set(itemsetx)), tidsetx)

    def _oneLengthFrequentItems(self):
        """
        To calculate the one Length items
        """
        Vector = {}
        items = []
        for i in self._Database:
            for j in self._plist:
                count = 0
                if j in i:
                    count = 1
                if j in Vector:
                    Vector[j].append(count)
                else:
                    Vector[j] = [count]
        for x, y in Vector.items():
            v = self._Count(y)
            if v >= self._itemSup:
                items.append(x)
        return Vector, items

    @deprecated(
        "It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """
        self.mine()

    def mine(self):
        """
        Frequent pattern mining process will start from here
        """
        self._Database = []
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._itemSup = self._convert(self._itemSup)
        self._minLength = int(self._minLength)
        self._faultTolerance = int(self._faultTolerance)
        Vector, plist = self._oneLengthFrequentItems()
        for i in range(len(plist)):
            itemx = plist[i]
            tidsetx = Vector[itemx]
            itemsetx = [itemx]
            itemsets = []
            tidsets = []
            for j in range(i + 1, len(plist)):
                itemj = plist[j]
                tidsetj = Vector[itemj]
                y1 = list(_np.array(tidsetx) | _np.array(tidsetj))
                total = self._Count(y1)
                if total >= self._minSup:
                    itemsets.append(itemj)
                    tidsets.append(y1)
            if (len(itemsets) > 0):
                self._processEquivalenceClass(itemsetx, itemsets, tidsets)
            self._save(None, itemsetx, tidsetx)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Fault-Tolerant Frequent patterns were generated successfully using VBFTMine algorithm ")

    def getMemoryUSS(self):
        """

        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """

        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """

        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """

        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            s = str()
            for i in a:
                s = s + i + ' '
            data.append([s, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # dataFrame = dataFrame.replace(r'\r+|\n+|\t+',' ', regex=True)
        return dataFrame

    def save(self, outFile):
        """

        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s = str()
            for i in x:
                s = s + i + '\t'
            s1 = s.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """

        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 7 or len(_ab._sys.argv) == 8:
        if len(_ab._sys.argv) == 8:
            _ap = VBFTMine(_ab._sys.argv[1], _ab._sys.argv[3],  _ab._sys.argv[4],
                            _ab._sys.argv[5], _ab._sys.argv[6], _ab._sys.argv[7],)
        if len(_ab._sys.argv) == 7:
            _ap = VBFTMine(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        _ap.mine()
        _ap.mine()
        print("Total number of Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        _ap = VBFTMine('/Users/Likhitha/Downloads/fault/sample4.txt', 5, 3, 2, 1, ' ')
        _ap.mine()
        _ap.printResults()
        print(_ap.getPatternsAsDataFrame())
        print("Error! The number of input parameters do not match the total number of parameters provided")



# DF2DB in this code dataframe is converting databases into sparse or dense transactional, temporal, Utility.
#
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.DF2DB import DF2DB as db
#
#             obj = db.DF2DB(idf, "sparse/dense")
#
#             obj.convert2Transactional("outputFileName", ">=", 16) # To create transactional database
#
#             obj.convert2Temporal("outputFileName", ">=", 16) # To create temporal database
#
#             obj.convert2Utility("outputFileName") # To create utility database
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import PAMI.extras.convert.denseDF2DB as dense
import PAMI.extras.convert.sparseDF2DB as sparse
import sys,psutil,os,time
from typing import Union
import operator

condition_operator = {
    '<': operator.lt,
    '>': operator.gt,
    '<=': operator.le,
    '>=': operator.ge,
    '==': operator.eq,
    '!=': operator.ne
}

class DF2DB:
    """
    :Description:  This class will create database for given DataFrame based on Threshold values and conditions are defined in the class.
                   Converts Dataframe into sparse or dense dataframes.

    :Attributes:

        :param inputDF: DataFrame :
             It is sparse or dense DataFrame
        :param thresholdValue: int or float :
             It is threshold value of all item
        :param condition: str :
             It is condition of all item
        :param DFtype: str :
             It is DataFrame type. It should be sparse or dense. Default DF is sparse.
        :param memoryUSS : float
            To store the total amount of USS memory consumed by the program
        :param memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        :param startTime : float
            To record the start time of the mining process
        endTime : float
            To record the completion time of the mining process


    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.DF2DB import DF2DB as db

            obj = db.DF2DB(idf, "sparse/dense")

            obj.convert2Transactional("outputFileName",condition,threshold) # To create transactional database

            obj.convert2Temporal("outputFileName",condition,threshold) # To create temporal database

            obj.convert2Utility("outputFileName",condition,threshold) # To create utility database
    """,


    def __init__(self, inputDF, DFtype='dense') -> None:
        self.inputDF = inputDF
        self.DFtype = DFtype.lower()
        if DFtype == 'sparse':
            self.DF2DB = sparse.sparseDF2DB(self.inputDF)
        elif DFtype == 'dense':
            self.DF2DB = dense.denseDF2DB(self.inputDF)
        else:
            raise Exception('DF type should be sparse or dense')
        self._startTime = float()
        self._endTime = float()
        self._memoryUSS = float()
        self._memoryRSS = float()
        self.tids = []
        self.items = []
        self.items = list(self.inputDF.columns.values)
        self.tids = list(self.inputDF.index)

    def convert2TransactionalDatabase(self, oFile: str, condition: str, thresholdValue: Union[int, float]) -> None:
        """
        create transactional database and return oFileName
        :param oFile: file name or path to store database
        :type oFile: str
        :return: oFile name
        :rtype: str
        """
        self._startTime = time.time()
        with open(oFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f'\t{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction[0]}')
                    else:
                        continue
                    f.write('\n')
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime  = time.time()

    def convert2TemporalDatabase(self, oFile: str, condition: str, thresholdValue: Union[int, float]) -> None:
        """
        create temporal database and return oFile name
        :param oFile: file name or path to store database
        :type oFile: str
        :return: oFile name
        :rtype: str
        """
        self._startTime = time.time()
        with open(oFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{tid + 1}')
                        for item in transaction:
                            f.write(f'\t{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid + 1}')
                        f.write(f'\t{transaction[0]}')
                    else:
                        continue
                    f.write('\n')
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime  = time.time()

    def convert2UtilityDatabase(self, oFile: str) ->None:
        """
        create utility database and return oFile name
        :param oFile:  file name or path to store database
        :type oFile: str
        :return: outputFile name
        :rtype: str
        """
        self._startTime = time.time()
        with open(oFile, 'w') as f:
            for tid in self.tids:
                df = self.inputDF.loc[tid].dropna()
                f.write(f'{df.index[0]}')
                for item in df.index[1:]:
                    f.write(f'\t{item}')
                f.write(f':{df.sum()}:')
                f.write(f'{df.at[df.index[0]]}')

                for item in df.index[1:]:
                    f.write(f'\t{df.at[item]}')
                f.write('\n')
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime = time.time()

    def convert2geoReferencedTransactionalDatabase(self, oFile: str, condition: str, thresholdValue: Union[int, float]) -> str:
        """
        create transactional database and return oFileName
        :param oFile: file name or path to store database
        :type oFile: str
        :return: oFile name
        :rtype: str
        """
        self._startTime = time.time()
        with open(oFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f'\t{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction[0]}')
                    else:
                        continue
                    f.write('\n')
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime  = time.time()
        return self.DF2DB.getFileName()

    def convert2geoReferencedTemporalDatabase(self, oFile: str, condition: str, thresholdValue: Union[int, float]) -> str:
        """
        create temporal database and return oFile name
        :param oFile: file name or path to store database
        :type oFile: str
        :return: oFile name
        :rtype: str
        """
        self._startTime = time.time()
        with open(oFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{tid + 1}')
                        for item in transaction:
                            f.write(f'\t{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid + 1}')
                        f.write(f'\t{transaction[0]}')
                    else:
                        continue
                    f.write('\n')
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime  = time.time()
        return self.DF2DB.getFileName()


    def convert2MultipleTimeSeries(self, oFile: str, condition: str,
                                   thresholdValue: Union[int, float], interval: int) -> None:
        """
        :Description: Create the multiple time series database.

        :param outputFile:  Write multiple time series database into outputFile.

        :type outputFile:  str

        :param interval: Breaks the given timeseries into intervals.

        :type interval: int

        :param condition: It is condition to judge the value in dataframe

        :param thresholdValue: User defined value.

        :type thresholdValue: int or float
        """
        self._startTime = time.time()
        writer = open(oFile, 'w+')
        # with open(self.outputFile, 'w+') as f:
        count = 0
        tids = []
        items = []
        values = []
        for tid in self.tids:
            count += 1
            transaction = [item for item in self.items if
                           condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
            for i in transaction:
                tids.append(count)
                items.append(i)
                values.append(self.inputDF.at[tid, i])
            if count == interval:
                s1, s, ss = str(), str(), str()
                if len(values) > 0:

                    for j in range(len(tids)):
                        s1 = s1 + str(tids[j]) + '\t'
                    for j in range(len(items)):
                        s = s + items[j] + '\t'
                    for j in range(len(values)):
                        ss = ss + str(values[j]) + '\t'

                s2 = s1 + ':' + s + ':' + ss
                writer.write("%s\n" % s2)
                tids, items, values = [], [], []
                count = 0
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime = time.time()



    def convert2UncertainTransactionalDatabase(self, oFile: str, condition: str,
                                       thresholdValue: Union[int, float]) -> None:
        with open(oFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    uncertain = [self.inputDF.at[tid, item] for item in self.items if
                                 condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f'\t{item}')
                        f.write(f':')
                        for value in uncertain:
                            tt = 0.1 + 0.036 * abs(25 - value)
                            tt = round(tt, 2)
                            f.write(f'\t{tt}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction[0]}')
                        tt = 0.1 + 0.036 * abs(25 - uncertain[0])
                        tt = round(tt, 2)
                        f.write(f':{tt}')
                    else:
                        continue
                    f.write('\n')


    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime



if __name__ == '__main__':
    obj = DF2DB(sys.argv[1])
    obj.getTransactionalDatabase(sys.argv[2],sys.argv[3],sys.argv[4])
    print("Conversion is complete.")
    print("Total Memory in USS:", obj.getMemoryUSS())
    print("Total Memory in RSS", obj.getMemoryRSS())
    print("Total ExecutionTime in ms:", obj.getRuntime())



import numpy as np
import pandas as pd
import random


class UtilityDataGenerator:
    def __init__(self, databaseSize, numberOfItems, averageLengthOfTransaction,
                 minimumInternalUtilityValue, maximumInternalUtilityValue,
                 minimumExternalUtilityValue, maximumExternalUtilityValue):
        self.databaseSize = databaseSize
        self.numberOfItems = numberOfItems
        self.averageLengthOfTransaction = averageLengthOfTransaction
        self.minInternalUtilityValue = minimumInternalUtilityValue
        self.maxInternalUtilityValue = maximumInternalUtilityValue
        self.minExternalUtilityValue = minimumExternalUtilityValue
        self.maxExternalUtilityValue = maximumExternalUtilityValue
        self.entries = []
        self.ExternalUtilityData = self.GenerateExternalUtilityData()

    def GenerateExternalUtilityData(self):
        items = range(1, self.numberOfItems + 1)
        ExternalUtilityData = {f'item{item}': random.randint(100, 900) for item in items}
        return ExternalUtilityData

    def Generate(self):
        for entry_id in range(1, self.databaseSize + 1):
            entry_length = np.random.randint(1, self.averageLengthOfTransaction * 2)
            entry = np.random.randint(self.minInternalUtilityValue, self.maxInternalUtilityValue + 1,
                                      size=self.numberOfItems)
            entry_sum = entry.sum()
            self.entries.append((entry, entry_sum))

    def Save(self, fileName):
        with open(fileName, 'w') as file:
            for idx, (entry, entry_sum) in enumerate(self.entries, start=1):
                entry_str = '\t'.join(map(str, entry))
                file.write(f'{idx}\t{entry_str}\t{entry_sum}\n')

    def SaveItemsInternalUtilityValues(self, fileName):
        items = random.sample(range(1, self.numberOfItems + 1), self.numberOfItems)
        internal_utility_data = [np.random.randint(self.minInternalUtilityValue, self.maxInternalUtilityValue + 1) for _
                                 in items]
        data = {'Item': items, 'Internal Utility Value': internal_utility_data}
        df = pd.DataFrame(data)
        df.to_csv(fileName, sep='\t', index=False)

    def Saveitemsexternalutilityvalues(self, fileName):
        items = random.sample(range(1, self.numberOfItems + 1), self.numberOfItems)
        data = {'Item': [f'item{item}' for item in items],
                'External Utility Value': list(self.ExternalUtilityData.values())}
        df = pd.DataFrame(data)
        df.to_csv(fileName, sep='\t', index=False)

    def GetUtilityData(self):
        data = {'Entry ID': range(1, len(self.entries) + 1),
                'Entries': [entry for entry, _ in self.entries],
                'Sum': [entry_sum for _, entry_sum in self.entries]}
        df = pd.DataFrame(data)
        return df

    def GetInternalUtilityData(self):
        items = random.sample(range(1, self.numberOfItems + 1), self.numberOfItems)
        InternalUtilityData = [np.random.randint(self.minInternalUtilityValue, self.maxInternalUtilityValue + 1) for _
                                 in items]
        data = {'Item': items, 'Internal Utility Value': InternalUtilityData}
        df = pd.DataFrame(data)
        return df

    def GetExternalUtilityData(self):
        items = random.sample(range(1, self.numberOfItems + 1), self.numberOfItems)
        data = {'Item': [f'item{item}' for item in items],
                'External Utility Value': list(self.ExternalUtilityData.values())}
        df = pd.DataFrame(data)
        return df

    def GenerateAndPrintItemPairs(self):
        items = random.sample(range(1, self.numberOfItems + 1), 2)
        item1_id = f'item{items[0]}'
        item2_id = f'item{items[1]}'
        item1_value = self.ExternalUtilityData[item1_id]
        item2_value = self.ExternalUtilityData[item2_id]
        sum_values = item1_value + item2_value
        print(f"{item1_id} value: {item1_value}\t{item2_id} value: {item2_value}\tSum of values: {sum_values}")

        # Separate the sum with ' : '
        print(f"{item1_value}:{item2_value}:{sum_values}")


if __name__ == "__main__":
    data_generator = UtilityDataGenerator(100000, 2000, 10, 1, 100, 1, 10)
    data_generator.Generate()
    data_generator.Save("utility_data-6.csv")
    data_generator.SaveItemsInternalUtilityValues("items_internal_utility.csv")
    data_generator.Saveitemsexternalutilityvalues("items_external_utility.csv")
    utility_data = data_generator.GetUtilityData()
    InternalUtilityData = data_generator.GetInternalUtilityData()
    ExternalUtilityData = data_generator.GetExternalUtilityData()

    for _ in range(10):  # Print pairs for demonstration, adjust the range as needed
        data_generator.GenerateAndPrintItemPairs()






















































#  Usage
#  obj1 = GraphConvertor('iFile', 'oFile')
#
#  obj1.convertTraditional2Compressed()
#
#  obj1.getMemoryRSS()
#
#  obj1.getMemoryUSS()
#
#  obj2 = GraphConvertor('iFileCompressed', 'oFileTrad')
#
#  obj2.convertCompressed2Traditional()
#
#  obj2.getMemoryRSS()
#
#  obj2.getMemoryUSS()

import os
import psutil


class GraphConvertor:
    def __init__(self, iFile):
        self.iFile = iFile
        self.convertedData = []

    def _writeGraphToFileCompressed(self, graph):
        node_str = ' '.join(f"{node} {label}" for node, label in sorted(graph['nodes'], key=lambda x: x[0]))
        edge_str = ' '.join(f"{u} {v} {label}" for u, v, label in graph['edges'])
        return f"{node_str} : {edge_str}\n"

    def _writeGraphToFileTraditional(self, graph, gId):
        traditional_lines = [f"t # {gId}\n"]
        for node, label in sorted(graph['nodes'], key=lambda x: x[0]):
            traditional_lines.append(f"v {node} {label}\n")
        for u, v, label in graph['edges']:
            traditional_lines.append(f"e {u} {v} {label}\n")
        return ''.join(traditional_lines)

    def convertTraditional2Compressed(self):
        graph = {}
        self.convertedData = []
        with open(self.iFile, 'r') as iFile:
            for line in iFile:
                parts = line.strip().split()
                if not parts:
                    continue
                if parts[0] == 't':
                    if graph:
                        compressedGraph = self._writeGraphToFileCompressed(graph)
                        self.convertedData.append(compressedGraph)
                    graph = {'nodes': [], 'edges': []}
                elif parts[0] == 'v':
                    graph['nodes'].append((int(parts[1]), parts[2]))
                elif parts[0] == 'e':
                    graph['edges'].append((int(parts[1]), int(parts[2]), parts[3]))
            if graph:
                compressedGraph = self._writeGraphToFileCompressed(graph)
                self.convertedData.append(compressedGraph)

    def convertCompressed2Traditional(self):
        self.convertedData = []
        gId = 0
        with open(self.iFile, 'r') as iFile:
            for line in iFile:
                if not line.strip():
                    continue  # Skip empty lines
                if ':' not in line:
                    print(f"Invalid format in line: {line.strip()}")
                    continue
                nodes_part, edges_part = line.strip().split(':')
                nodes_tokens = nodes_part.strip().split()
                edges_tokens = edges_part.strip().split()

                # Parse nodes
                nodes = []
                for i in range(0, len(nodes_tokens), 2):
                    node_id = int(nodes_tokens[i])
                    node_label = nodes_tokens[i + 1]
                    nodes.append((node_id, node_label))

                # Parse edges
                edges = []
                for i in range(0, len(edges_tokens), 3):
                    if i + 2 >= len(edges_tokens):
                        print(f"Incomplete edge information in line: {line.strip()}")
                        break
                    u = int(edges_tokens[i])
                    v = int(edges_tokens[i + 1])
                    label = edges_tokens[i + 2]
                    edges.append((u, v, label))

                graph = {'nodes': nodes, 'edges': edges}
                traditionalGraph = self._writeGraphToFileTraditional(graph, gId)
                self.convertedData.append(traditionalGraph)
                gId += 1

    def save(self, oFile):
        """
        Saves the converted data to the specified output file.

        :param oFile: Path to the output file.
        """
        if not self.convertedData:
            print("No converted data to save. Please perform a conversion first.")
            return

        with open(oFile, 'w') as file:
            for graphData in self.convertedData:
                file.write(graphData)

    def getMemoryRSS(self):
        process = psutil.Process(os.getpid())
        rss = process.memory_info().rss
        return rss

    def getMemoryUSS(self):
        process = psutil.Process(os.getpid())
        uss = process.memory_full_info().uss
        return uss




# plotLineGraphsFromDataFrame is used to convert the given dataframe into a Line Graph.
#
#  **Importing this algorithm into a python program**
#
#     from PAMI.extras.graph import plotLineGraphsFromDataFrame as plt
#
#     dataFrame = pd.DataFrame(data)
#
#     obj = plt.plotLineGraphsFromDataFrame(dataFrame)
#
#     obj.plot(result=dataFrame, xaxis='minSup', yaxis='patterns', label='algorithm')
#
#     obj.plot(result=dataFrame, xaxis='minSup', yaxis='runtime', label='algorithm')
#
#     obj.plot(result=dataFrame, xaxis='minSup', yaxis='memoryRSS', label='algorithm')
#
#     obj.plot(result=dataFrame, xaxis='minSup', yaxis='memoryUSS', label='algorithm')
#
#     obj.save(result=dataFrame, xaxis='minSup', yaxis='patterns', label='algorithm', oFile='patterns.png')
#
#     obj.save(result=dataFrame, xaxis='minSup', yaxis='runtime', label='algorithm', oFile='runtime.png')
#
#     obj.save(result=dataFrame, xaxis='minSup', yaxis='memoryRSS', label='algorithm', oFile='memoryRSS.png')
#
#     obj.save(result=dataFrame, xaxis='minSup', yaxis='memoryUSS', label='algorithm', oFile='memoryUSS.png')
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import matplotlib.pyplot as plt
import pandas as pd


class plotLineGraphsFromDataFrame:
    """
    A class to generate and save line graphs from a given DataFrame.

    :Attributes:
        dataFrame (pd.DataFrame): Input DataFrame to generate graphs.

    :Methods:
        plot(result, xaxis, yaxis, label): Plots a line graph based on the specified axes.
        save(result, xaxis, yaxis, label, oFile): Saves the line graph to the specified file.
    """

    def __init__(self, dataFrame: pd.DataFrame) -> None:
        """
        Initialize the class with a DataFrame.

        :param dataFrame: Input DataFrame
        """
        self.dataFrame = dataFrame

    def plot(self, result: pd.DataFrame, xaxis: str, yaxis: str, label: str) -> None:
        """
        Plots a line graph.

        :param result: Input DataFrame
        :param xaxis: Column name for the x-axis
        :param yaxis: Column name for the y-axis
        :param label: Column name to use for legend labels
        """
        plt.figure()
        for key, grp in result.groupby(label):
            plt.plot(grp[xaxis], grp[yaxis], label=f"{label}: {key}")
        plt.xlabel(xaxis)
        plt.ylabel(yaxis)
        plt.title(f"{yaxis} vs {xaxis}")
        plt.legend()
        plt.show()
        print(f"Graph for {yaxis} vs {xaxis} is successfully generated!")

    def save(self, result: pd.DataFrame, xaxis: str, yaxis: str, label: str, oFile: str) -> None:
        """
        Saves the line graph to a file.

        :param result: Input DataFrame
        :param xaxis: Column name for the x-axis
        :param yaxis: Column name for the y-axis
        :param label: Column name to use for legend labels
        :param oFile: Output file name to save the graph
        """
        plt.figure()
        for key, grp in result.groupby(label):
            plt.plot(grp[xaxis], grp[yaxis], label=f"{label}: {key}")
        plt.xlabel(xaxis)
        plt.ylabel(yaxis)
        plt.title(f"{yaxis} vs {xaxis}")
        plt.legend()
        plt.savefig(oFile)
        plt.close()
        print(f"Graph saved as {oFile}!")

if __name__ == "__main__":
    # Example DataFrame
    data = {
        'algorithm': ['FPGrowth', 'FPGrowth', 'FPGrowth', 'FPGrowth', 'FPGrowth'],
        'minSup': [0.01, 0.02, 0.03, 0.01, 0.02],
        'patterns': [386, 155, 60, 386, 155],
        'runtime': [7.35, 4.66, 4.66, 4.57, 2.51],
        'memoryRSS': [426545152, 309182464, 241397760, 233537536, 267165696],
        'memoryUSS': [426545152, 309182464, 241397760, 233537536, 267165696]
    }
    dataFrame = pd.DataFrame(data)

    obj = plotLineGraphsFromDataFrame(dataFrame)

    obj.plot(result=dataFrame, xaxis='minSup', yaxis='patterns', label='algorithm')
    obj.plot(result=dataFrame, xaxis='minSup', yaxis='runtime', label='algorithm')
    obj.plot(result=dataFrame, xaxis='minSup', yaxis='memoryRSS', label='algorithm')
    obj.plot(result=dataFrame, xaxis='minSup', yaxis='memoryUSS', label='algorithm')

    obj.save(result=dataFrame, xaxis='minSup', yaxis='patterns', label='algorithm', oFile='patterns.png')
    obj.save(result=dataFrame, xaxis='minSup', yaxis='runtime', label='algorithm', oFile='runtime.png')
    obj.save(result=dataFrame, xaxis='minSup', yaxis='memoryRSS', label='algorithm', oFile='memoryRSS.png')
    obj.save(result=dataFrame, xaxis='minSup', yaxis='memoryUSS', label='algorithm', oFile='memoryUSS.png')



# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             import PAMI.periodicFrequentPattern.topk.TopkPFPGrowth as alg
#
#             obj = alg.TopkPFPGrowth(iFile, k, maxPer,oFile)
#
#             obj.mine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.periodicFrequentPattern.basic import abstract as _ab
import pandas as pd
from deprecated import deprecated


class TopkPFPGrowth(_ab._periodicFrequentPatterns):
    """
    :Description:   Top - K is and algorithm to discover top periodic frequent patterns in a temporal database.

    :Reference:   Komate Amphawan, Philippe Lenca, Athasit Surarerks: "Mining Top-K Periodic-Frequent Pattern from Transactional Databases without Support Threshold"
                  International Conference on Advances in Information Technology: https://link.springer.com/chapter/10.1007/978-3-642-10392-6_3

    :param  iFile: str :
                   Name of the Input file to mine complete set of periodic frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of periodic frequent pattern's
    :param  maxPer: str:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        iFile : str
            Input file name or path of the input file
        k: int
            User specified counte of top frequent patterns
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        oFile : str
            Name of the output file or the path of the output file
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets()
            Scans the dataset or dataframes and stores in list format
        frequentOneItem()
            Generates one frequent patterns
        eclatGeneration(candidateList)
            It will generate the combinations of frequent items
        generateFrequentPatterns(tidList)
            It will generate the combinations of frequent items from a list of items

    **Executing the code on terminal:**
    -------------------------------------
   .. code-block:: console


       Format:

       (.venv) $ python3 TopkPFP.py <inputFile> <outputFile> <k> <maxPer>

       Examples:

       (.venv) $ python3 TopkPFP.py sampleDB.txt patterns.txt 10 3


    **Sample run of the importing code:**
    ---------------------------------------
    .. code-block:: python

            import PAMI.periodicFrequentPattern.topk.TopkPFPGrowth as alg

            obj = alg.TopkPFPGrowth(iFile, k, maxPer)

            obj.mine()

            periodicFrequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(periodicFrequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    --------------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

    """

    _startTime = float()
    _endTime = float()
    _k = int()
    _maxPer = " "
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _tidList = {}
    _lno = int()
    _minimum = int()
    _mapSupport = {}

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable

        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            if 'Patterns' in i:
                data = self._iFile['Patterns'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                self._Database.append(tr)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _frequentOneItem(self):
        """
        Generating one frequent patterns
        """

        self._mapSupport = {}
        self._tidList = {}
        n = 0
        for line in self._Database:
            self._lno += 1
            n = int(line[0])
            for i in range(1, len(line)):
                si = line[i]
                if self._mapSupport.get(si) is None:
                    self._mapSupport[si] = [1, abs(0 - n), n]
                    self._tidList[si] = [n]
                else:
                    self._mapSupport[si][0] += 1
                    self._mapSupport[si][1] = max(self._mapSupport[si][1], abs(n - self._mapSupport[si][2]))
                    self._mapSupport[si][2] = n
                    self._tidList[si].append(n)
        for x, y in self._mapSupport.items():
            self._mapSupport[x][1] = max(self._mapSupport[x][1], abs(n - self._mapSupport[x][2]))
        self._maxPer = self._convert(self._maxPer)
        self._k = self._convert(self._k)
        self._mapSupport = {k: [v[0], v[1]] for k, v in self._mapSupport.items() if v[1] <= self._maxPer}
        plist = [key for key, value in sorted(self._mapSupport.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        self._finalPatterns = {}
        #print(len(plist))
        for i in plist:
            if len(self._finalPatterns) >= self._k:
                break
            else:
                self._finalPatterns[i] = [self._mapSupport[i][0], self._mapSupport[i][1]]
        self._minimum = min([self._finalPatterns[i][0] for i in self._finalPatterns.keys()])
        plist = list(self._finalPatterns.keys())
        return plist

    def _getSupportAndPeriod(self, timeStamps):
        """To calculate the periodicity and support

        :param timeStamps: Timestamps of an item set
        :return: support, periodicity
        """

        global lno
        timeStamps.sort()
        cur = 0
        per = list()
        sup = 0
        for j in range(len(timeStamps)):
            per.append(timeStamps[j] - cur)
            cur = timeStamps[j]
            sup += 1
        per.append(self._lno - cur)
        if len(per) == 0:
            return [0, 0]
        return [sup, max(per)]

    def _save(self, prefix, suffix, tidSetI):
        """Saves the patterns that satisfy the periodic frequent property.

        :param prefix: the prefix of a pattern
        :type prefix: list
        :param suffix: the suffix of a patterns
        :type suffix: list
        :param tidSetI: the timestamp of a patterns
        :type tidSetI: list
        """

        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        val = self._getSupportAndPeriod(tidSetI)
        sample = str()
        for i in prefix:
            sample = sample + i + " "
        if len(self._finalPatterns) < self._k:
            if val[0] >= self._minimum:
                self._finalPatterns[sample] = val
                self._finalPatterns = {k: v for k, v in
                                  sorted(self._finalPatterns.items(), key=lambda item: item[1], reverse=True)}
                self._minimum = min([self._finalPatterns[i][0] for i in self._finalPatterns.keys()])
        else:
            for x, y in sorted(self._finalPatterns.items(), key=lambda x: x[1][0]):
                if val[0] > y[0]:
                    del self._finalPatterns[x]
                    self._finalPatterns[x] = y
                    self._finalPatterns = {k: v for k, v in
                                          sorted(self._finalPatterns.items(), key=lambda item: item[1], reverse=True)}
                    self._minimum = min([self._finalPatterns[i][0] for i in self._finalPatterns.keys()])
                    return

    def _Generation(self, prefix, itemSets, tidSets):
        """
        Equivalence class is followed  and checks for the patterns generated for periodic-frequent patterns.

        :param prefix:  main equivalence prefix
        :type prefix: periodic-frequent item or pattern
        :param itemSets: patterns which are items combined with prefix and satisfying the periodicity
                            and frequent with their timestamps
        :type itemSets: list
        :param tidSets: timestamps of the items in the argument itemSets
        :type tidSets: list
        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidI = tidSets[0]
            self._save(prefix, [i], tidI)
            return
        for i in range(len(itemSets)):
            itemI = itemSets[i]
            if itemI is None:
                continue
            tidSetI = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemI]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                tidSetJ = tidSets[j]
                y = list(set(tidSetI).intersection(tidSetJ))
                val = self._getSupportAndPeriod(y)
                if val[0] >= self._minimum and val[1] <= self._maxPer:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            newPrefix = list(set(itemSetX)) + prefix
            self._Generation(newPrefix, classItemSets, classTidSets)
            self._save(prefix, list(set(itemSetX)), tidSetI)

    @deprecated("It is recommended to use mine() instead of mine() for mining process")
    def startMine(self):
        """
        Main function of the program
        """
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._k is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        _plist = self._frequentOneItem()
        for i in range(len(_plist)):
            itemI = _plist[i]
            tidSetI = self._tidList[itemI]
            itemSetX = [itemI]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(_plist)):
                itemJ = _plist[j]
                tidSetJ = self._tidList[itemJ]
                y1 = list(set(tidSetI).intersection(tidSetJ))
                val = self._getSupportAndPeriod(y1)
                if val[0] >= self._minimum and val[1] <= self._maxPer:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self._Generation(itemSetX, itemSets, tidSets)
        print("TopK Periodic Frequent patterns were generated successfully")
        self._endTime = _ab._time.time()
        _process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = _process.memory_full_info().uss
        self._memoryRSS = _process.memory_info().rss

    def Mine(self):
        """
        Main function of the program
        """
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._k is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        _plist = self._frequentOneItem()
        for i in range(len(_plist)):
            itemI = _plist[i]
            tidSetI = self._tidList[itemI]
            itemSetX = [itemI]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(_plist)):
                itemJ = _plist[j]
                tidSetJ = self._tidList[itemJ]
                y1 = list(set(tidSetI).intersection(tidSetJ))
                val = self._getSupportAndPeriod(y1)
                if val[0] >= self._minimum and val[1] <= self._maxPer:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self._Generation(itemSetX, itemSets, tidSets)
        print("TopK Periodic Frequent patterns were generated successfully")
        self._endTime = _ab._time.time()
        _process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = _process.memory_full_info().uss
        self._memoryRSS = _process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.replace(' ', '\t') + ":" + f'{y[0]}:{y[1]}'
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Top K Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = TopkPFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = TopkPFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.mine()
        print("Top K Periodic-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:",  _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



# TransactionalDatabase is a collection of transactions. It only considers the data in  transactions and ignores the metadata.
#
#  **Importing this algorithm into a python program**
#  --------------------------------------------------------
#     from PAMI.extras.syntheticDataGenerator import TransactionalDatabase as db
#
#     obj = db(10, 5, 10)
#
#     obj.create()
#
#     obj.save('db.txt')
#
#     print(obj.getTransactions())
#
import numpy as np
import pandas as pd
import sys, psutil, os, time

__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


class TransactionalDatabase:
    """
        :Description: TransactionalDatabase is a collection of transactions. It only considers the data in  transactions and ignores the metadata.
        :Attributes:

            dataBaseSize: int
                Number of Transactions in a database
            avgItemsPerTransaction: int
                Average number of items per transaction
            itemsNo: int
                Total number of items
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                        To store the total amount of RSS memory consumed by the program
            startTime : float
                        To record the start time of the mining process
            endTime : float
                        To record the completion time of the mining process

        :Methods:

            create:
                Generate the transactional database
            save:
                Save the transactional database to a user-specified file
            getTransactions:
                Get the transactional database
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function

        **Methods to execute code on terminal**
        ---------------------------------------------

        .. code-block:: console

          Format:

          (.venv) $ python3 TransactionalDatabase.py <dataBaseSize> <avgItemsPerTransaction> <itemsNo>

          Example Usage:

          (.venv) $ python3 TransactionalDatabase.py 50.0 10.0 100


        **Importing this algorithm into a python program**
        --------------------------------------------------------
            from PAMI.extras.syntheticDataGenerator import TransactionalDatabase as db

            obj = db.TransactionalDatabase(10, 5, 10)

            obj.create()

            obj.save('db.txt')

            print(obj.getTransactions())


        """

    def __init__(self, databaseSize, avgItemsPerTransaction, numItems,sep = "\t") -> None:

        """
        Initialize the transactional database with the given parameters

        :param databaseSize: total number of transactions in the database
        :type databaseSize: int
        :param avgItemsPerTransaction: average number of items per transaction
        :type avgItemsPerTransaction: int
        :param numItems: total number of items
        :type numItems: int
        :param sep: separator to distinguish the items in a transaction
        :type sep: str
        """

        self.databaseSize = databaseSize
        self.avgItemsPerTransaction = avgItemsPerTransaction
        self.numItems = numItems
        self.sep = sep
        self.db = []
        self._startTime = float()
        self._endTime = float()
        self._memoryUSS = float()
        self._memoryRSS = float()


    def _generateArray(self,databaseSize,avgItemsPerTransaction):

        #generating random numbers with size of dataBaseSize
        transactionSize = np.random.rand(databaseSize)

        #sum of the values in the transactionSize array - sumTransactions
        sumTransactions = np.sum(transactionSize)

        #weights of the values in the array - transactionSize
        weights = transactionSize/sumTransactions

        #sumResultant -> whose average is equal to averageItemsPerTransaction's value
        sumResultant = avgItemsPerTransaction*databaseSize

        #Getting new values whose average is averageItemsPerTransaction
        newValues = np.round(weights*sumResultant)

        #changing the values in numpy array to int
        valuesInt = newValues.astype(int)

        #finding if there are any 0's in the array
        indexZero = np.where(valuesInt==0)[0]

        #Adding +1 to the transactions which have 0
        if len(indexZero)==0:
            return valuesInt
        else:
            for i in indexZero:
                valuesInt[i]+=1

        return valuesInt

    def create(self):
        self._startTime = time.time()
        noofItemsperTrans = self._generateArray(self.databaseSize, self.avgItemsPerTransaction)
        for i in range(self.databaseSize):
            self.db.append(np.random.choice(range(1,self.numItems+1), noofItemsperTrans[i], replace=False))

        self._endTime = time.time()

    def save(self, filename):
        with open(filename, 'w') as f:
            for line in self.db:
                f.write(str(self.sep).join(map(str, line)) + '\n')


    def getTransactions(self,sep='\t'):
        transactions = []
        for line in self.db:
            transactions.append(line)

        dataFrame = pd.DataFrame([transactions], index=['Transactions']).T
        return dataFrame


    def getMemoryUSS(self) -> float:

        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        return self._memoryUSS

    def getMemoryRSS(self) -> float:

        process = psutil.Process(os.getpid())
        self._memoryRSS = process.memory_info().rss
        return self._memoryRSS

    def getRuntime(self) -> float:
        return self._endTime - self._startTime

if __name__ == "__main__":

    if len(sys.argv) == 5:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]))
        obj.create()
        obj.save(sys.argv[4])
        print("Total Memory in USS:", obj.getMemoryUSS())
        print("Total Memory in RSS", obj.getMemoryRSS())
        print("Total ExecutionTime in ms:", obj.getRuntime())
    elif len(sys.argv) == 6:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]), sys.argv[4])
        obj.create()
        obj.save(sys.argv[5])
        print("Total Memory in USS:", obj.getMemoryUSS())
        print("Total Memory in RSS", obj.getMemoryRSS())
        print("Total ExecutionTime in ms:", obj.getRuntime())
    elif len(sys.argv) == 4:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]))
        obj.create()
        print("Total Memory in USS:", obj.getMemoryUSS())
        print("Total Memory in RSS", obj.getMemoryRSS())
        print("Total ExecutionTime in ms:", obj.getRuntime())
    else:
        raise ValueError(
            "Invalid number of arguments. Args: <databaseSize> <avgItemsPerTransaction> <noOfItems> <filename> or Args: <databaseSize> <avgItemsPerTransaction> <noOfItems> <sep> <filename>")




# Stable periodic pattern mining aims to discover all interesting patterns in a temporal database using three constraints minimum support,
# maximum period and maximum liability, that have support no less than the user-specified minimum support  constraint and liability no
# greater than maximum liability.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.stablePeriodicFrequentPattern.basic import SPPGrowth as alg
#
#             obj = alg.SPPGrowth(iFile, minSup, maxPer, maxLa)
#
#             obj.mine()
#
#             Patterns = obj.getPatterns()
#
#             print("Total number of Stable Periodic Frequent Patterns:", len(Patterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#




__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""


from PAMI.stablePeriodicFrequentPattern.basic import abstract as _ab
from deprecated import deprecated


_minSup = int()
_maxPer = int()
_maxLa = int()
_last = int()


class _Node:

    def __init__(self, item, children):
        """
        Initializing the Node class

        :param item: Storing the item of a node
        :type item: int or None
        :param children: To maintain the children of a node
        :type children: dict
        """

        self.item = item
        self.children = children
        self.parent = None
        self.timeStamps = []

    def addChild(self, node):
        """
        To add the children to a node

        :param node: parent node in the tree
        """

        self.children[node.item] = node
        node.parent = self

class _Tree:
    def __init__(self):
        self.root = _Node(None, {})
        self.summaries = {}
        self.info = {}

    def addTransaction(self, transaction, tid):
        """
        Adding a transaction into tree

        :param transaction: To represent the complete database
        :type transaction: list
        :param tid: To represent the timestamp of a database
        :type tid: list
        :return: pfp-growth tree
        """

        currentNode = self.root
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
        currentNode.timeStamps = currentNode.timeStamps + tid

    def getConditionalPatterns(self, alpha):
        """
        Generates all the conditional patterns of a respective node

        :param alpha: To represent a Node in the tree
        :type alpha: Node
        :return: A tuple consisting of finalPatterns, conditional pattern base and information
        """
        finalPatterns = []
        finalSets = []
        for i in self.summaries[alpha]:
            set1 = i.timeStamps
            set2 = []
            while i.parent.item is not None:
                set2.append(i.parent.item)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalSets.append(set1)
        finalPatterns, finalSets, info = self.conditionalDatabases(finalPatterns, finalSets)
        return finalPatterns, finalSets, info

    @staticmethod
    def generateTimeStamps(node):
        """
        To get the timestamps of a node

        :param node: A node in the tree
        :return: Timestamps of a node
        """

        finalTimeStamps = node.timeStamps
        return finalTimeStamps

    def removeNode(self, nodeValue):
        """
        Removing the node from tree

        :param nodeValue: To represent a node in the tree
        :type nodeValue: node
        :return: Tree with their nodes updated with timestamps
        """

        for i in self.summaries[nodeValue]:
            i.parent.timeStamps = i.parent.timeStamps + i.timeStamps
            del i.parent.children[nodeValue]

    def getTimeStamps(self, alpha):
        """
        To get all the timestamps of the nodes which share same item name

        :param alpha: Node in a tree
        :return: Timestamps of a  node
        """
        temporary = []
        for i in self.summaries[alpha]:
            temporary += i.timeStamps
        return temporary

    @staticmethod
    def getSupportAndPeriod(timeStamps):
        """
        To calculate the periodicity and support

        :param timeStamps: Timestamps of an item set
        :return: support, periodicity
        """
        global _maxPer, _last
        previous = 0
        la = 0
        tsList = sorted(timeStamps)
        laList = []
        for ts in tsList:
            la = max(0, la + ts - previous - _maxPer)
            laList.append(la)
            previous = ts
        la = max(0, la + _last - previous - _maxPer)
        laList.append(la)

        maxla = max(laList)
        return len(timeStamps), maxla

    def conditionalDatabases(self, conditionalPatterns, conditionalTimeStamps):
        """
        It generates the conditional patterns with periodic-frequent items

        :param conditionalPatterns: conditionalPatterns generated from conditionPattern method of a respective node
        :type conditionalPatterns: list
        :param conditionalTimeStamps: Represents the timestamps of a conditional patterns of a node
        :type conditionalTimeStamps: list
        :returns: Returns conditional transactions by removing non-periodic and non-frequent items
        """

        global _maxPer, _minSup, _maxLa
        pat = []
        timeStamps = []
        data1 = {}
        for i in range(len(conditionalPatterns)):
            for j in conditionalPatterns[i]:
                if j in data1:
                    data1[j] = data1[j] + conditionalTimeStamps[i]
                else:
                    data1[j] = conditionalTimeStamps[i]
        updatedDictionary = {}
        for m in data1:
            updatedDictionary[m] = self.getSupportAndPeriod(data1[m])
        updatedDictionary = {k: v for k, v in updatedDictionary.items() if v[0] >= _minSup and v[1] <= _maxLa}
        count = 0
        for p in conditionalPatterns:
            p1 = [v for v in p if v in updatedDictionary]
            trans = sorted(p1, key=lambda x: (updatedDictionary.get(x)[0], -x), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                timeStamps.append(conditionalTimeStamps[count])
            count += 1
        return pat, timeStamps, updatedDictionary

    def generatePatterns(self, prefix):
        """
        Generates the patterns

        :param prefix: Forms the combination of items
        :type prefix: list
        :returns: yields patterns with their support and periodicity
        """

        for i in sorted(self.summaries, key=lambda x: (self.info.get(x)[0], -x)):
            pattern = prefix[:]
            pattern.append(i)
            yield pattern, self.info[i]
            patterns, timeStamps, info = self.getConditionalPatterns(i)
            conditionalTree = _Tree()
            conditionalTree.info = info.copy()
            for pat in range(len(patterns)):
                conditionalTree.addTransaction(patterns[pat], timeStamps[pat])
            if len(patterns) > 0:
                for q in conditionalTree.generatePatterns(pattern):
                    yield q
            self.removeNode(i)

class SPPGrowth():
    """
    :Description:   Stable periodic pattern mining aims to dicover all interesting patterns in a temporal database using three contraints minimum support,
                    maximum period and maximum lability, that have support no less than the user-specified minimum support  constraint and lability no
                    greater than maximum lability.

    :Reference:   Dao, H.N. et al. (2022). Towards Efficient Discovery of Stable Periodic Patterns in Big Columnar Temporal Databases.
                  In: Fujita, H., Fournier-Viger, P., Ali, M., Wang, Y. (eds) Advances and Trends in Artificial Intelligence.
                  Theory and Practices in Artificial Intelligence. IEA/AIE 2022. Lecture Notes in Computer Science(), vol 13343. Springer, Cham.
                  https://doi.org/10.1007/978-3-031-08530-7_70

    :param  iFile: str :

                   Name of the Input file to mine complete set of frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minSup: str:
                   Minimum number of frequent patterns to be included in the output file.
    :param  maxLa: float:
                   Minimum number of ...
    :param  maxPer: float:
                   Maximum number of frequent patterns to be included in the output file.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

            iFile : file
                Name of the Input file or path of the input file
            oFile : file
                Name of the output file or path of the output file
            minSup: int or float or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            maxPer : int or float or str
                The user can specify maxPer either in count or proportion of database size.
                If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
                Otherwise, it will be treated as float.
                Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
            maxLa : int or float or str
                The user can specify maxLa either in count or proportion of database size.
                If the program detects the data type of maxLa is integer, then it treats maxLa is expressed in count.
                Otherwise, it will be treated as float.
                Example: maxLa=10 will be treated as integer, while maxLa=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            startTime : float
                To record the start time of the mining process
            endTime : float
                To record the completion time of the mining process
            Database : list
                To store the transactions of a database in list
            mapSupport : Dictionary
                To maintain the information of item and their frequency
            lno : int
                To represent the total no of transaction
            tree : class
                To represents the Tree class
            itemSetCount : int
                To represents the total no of patterns
            finalPatterns : dict
                To store the complete patterns

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of periodic-frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of periodic-frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets(fileName)
            Scans the dataset and stores in a list format
        PeriodicFrequentOneItem()
            Extracts the one-periodic-frequent patterns from database
        updateDatabases()
            Update the database by removing aperiodic items and sort the Database by item decreased support
        buildTree()
            After updating the Database, remaining items will be added into the tree by setting root node as null
        convert()
            to convert the user specified value

    **Methods to execute code on terminal**
    -----------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 topk.py <inputFile> <outputFile> <minSup> <maxPer> <maxLa>

      Example usage :

      (.venv) $ python3 topk.py sampleTDB.txt patterns.txt 0.3 0.4 0.3

    .. note:: constraints will be considered in percentage of database transactions

    **Importing this algorithm into a python program**
    -----------------------------------------------------
    .. code-block:: python

            from PAMI.stablePeriodicFrequentPattern.basic import topk as alg

            obj = alg.topk(iFile, minSup, maxPer, maxLa)

            obj.mine()

            Patterns = obj.getPatterns()

            print("Total number of Stable Periodic Frequent Patterns:", len(Patterns))

            obj.save(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    --------------

            The complete program was written by  P.Likhitha under the supervision of Professor Rage Uday Kiran.

    """
    _startTime = float()
    _endTime = float()
    _minSup = str()
    _maxPer = float()
    _maxLa = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankedUp = {}
    _lno = 0
    SPPList = {}

    def __init__(self, inputFile, minSup, maxPer, maxLa, sep='\t'):
        self._iFile = inputFile
        self._minSup = minSup
        self._maxPer = maxPer
        self._maxLa = maxLa
        self._sep = sep

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                self._Database.append(tr)

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _periodicFrequentOneItem(self):
        """
        Calculates the support of each item in the database and assign ranks to the items by decreasing support and returns the frequent items list

        :returns: return the one-length periodic frequent patterns
        """
        global _last
        tidLast = {}
        la = {}
        self.SPPList = {}
        for transaction in self._Database:
            ts = int(transaction[0])
            for item in transaction[1:]:
                if item not in self.SPPList:
                    la[item] = max(0, ts - self._maxPer)
                    self.SPPList[item] = [1, la[item]]
                else:
                    s = self.SPPList[item][0] + 1
                    la[item] = max(0, la[item] + ts - tidLast.get(item) - self._maxPer)
                    self.SPPList[item] = [s, max(la[item], self.SPPList[item][1])]
                tidLast[item] = ts
            _last = ts
        for item in self.SPPList:
            la[item] = max(0, la[item] + _last - tidLast[item] - self._maxPer)
            self.SPPList[item][1] = max(la[item], self.SPPList[item][1])
        self.SPPList = {k: v for k, v in self.SPPList.items() if v[0] >= self._minSup and v[1] <= self._maxLa}
        self.SPPList = {k: v for k, v in sorted(self.SPPList.items(), key=lambda x: x[1][0], reverse=True)}
        data = self.SPPList
        pfList = [k for k, v in sorted(data.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        self._rank = dict([(index, item) for (item, index) in enumerate(pfList)])
        #print(len(pfList))
        return data, pfList

    def _updateDatabases(self, dict1):
        """
        Remove the items which are not frequent from database and updates the database with rank of items

        :param dict1: frequent items with support
        :type dict1: dictionary
        :return: Sorted and updated transactions
        """
        list1 = []
        for tr in self._Database:
            list2 = [int(tr[0])]
            for i in range(1, len(tr)):
                if tr[i] in dict1:
                    list2.append(self._rank[tr[i]])
            if len(list2) >= 2:
                basket = list2[1:]
                basket.sort()
                list2[1:] = basket[0:]
                list1.append(list2)
        return list1

    @staticmethod
    def _buildTree(data, info):
        """
        It takes the database and support of each item and construct the main tree by setting root node as a null

        :param data: it represents the one Database in database
        :type data: list
        :param info: it represents the support of each item
        :type info: dictionary
        :return: returns root node of tree
        """

        rootNode = _Tree()
        rootNode.info = info.copy()
        for i in range(len(data)):
            set1 = [data[i][0]]
            rootNode.addTransaction(data[i][1:], set1)
        return rootNode

    def _savePeriodic(self, itemSet):
        """
        To convert the ranks of items in to their original item names

        :param itemSet: frequent pattern.
        :return: frequent pattern with original item names
        """
        t1 = str()
        for i in itemSet:
            t1 = t1 + self._rankedUp[i] + "\t"
        return t1

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    @deprecated("It is recommended to use mine() instead of mine() for mining process")
    def startMine(self):
        """
        Mining process will start from this function
        """

        self.mine()

    def mine(self):
        """
        Mining process will start from this function
        """

        global _minSup, _maxPer, _lno, _maxLa
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minSup is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        self._maxLa = self._convert(self._maxLa)
        _minSup, _maxPer, _maxLa, _lno = self._minSup, self._maxPer, self._maxLa, len(self._Database)
        #print(_minSup, _maxPer, _maxLa)
        if self._minSup > len(self._Database):
            raise Exception("Please enter the minSup in range between 0 to 1")
        generatedItems, pfList = self._periodicFrequentOneItem()
        updatedDatabases = self._updateDatabases(generatedItems)
        for x, y in self._rank.items():
            self._rankedUp[y] = x
        info = {self._rank[k]: v for k, v in generatedItems.items()}
        Tree = self._buildTree(updatedDatabases, info)
        patterns = Tree.generatePatterns([])
        self._finalPatterns = {}
        for i in patterns:
            sample = self._savePeriodic(i[0])
            self._finalPatterns[sample] = i[1]
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Stable Periodic Frequent patterns were generated successfully using topk algorithm ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of periodic-frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of periodic-frequent patterns after completion of the mining process

        :return: returning periodic-frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Stable Periodic  Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 6 or len(_ab._sys.argv) == 7:
        if len(_ab._sys.argv) == 7:
            _ap = SPPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        if len(_ab._sys.argv) == 6:
            _ap = SPPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        _ap.mine()
        _ap.mine()
        print("Total number of Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")




#
# obj = SyntheticGraphGenerator(numOfGraphTransactions, avgNumVertices, avgNumEdges, distinctVertexlLabels, distinctEdgeLabels, 'oFile', graphFormat)
# graphFormat = 'old' or 'new', default is 'old'
#
import random
from PAMI.extras.visualize import graphs

class SyntheticGraphGenerator:
    def __init__(self, numGraphs, avgNumVertices, avgNumEdges, numVertexLabels, numEdgeLabels, outputFileName, format='old'):
        self.numGraphs = numGraphs
        self.avgNumVertices = avgNumVertices
        self.avgNumEdges = avgNumEdges
        self.numVertexLabels = numVertexLabels
        self.numEdgeLabels = numEdgeLabels
        self.outputFileName = outputFileName
        self.format = format
        self._validate()
        self.generate()

    def _validate(self):
        if self.avgNumVertices < 1:
            raise ValueError("Average number of vertices should be greater than 0")
        if self.avgNumEdges < 0:
            raise ValueError("Average number of edges should be greater than or equal to 0")
        if self.numVertexLabels < 1:
            raise ValueError("Number of labels should be greater than 0")
        if self.numEdgeLabels < 1:
            raise ValueError("Number of labels should be greater than 0")
        if self.numGraphs < 1:
            raise ValueError("Number of graphs should be greater than 0")

        if self.avgNumVertices < self.avgNumEdges:
            raise ValueError("Average number of vertices should be greater than or equal to average number of edges")

        if self.avgNumEdges > self.avgNumVertices * (self.avgNumVertices - 1) / 2:
            raise ValueError("Average number of edges should be less than or equal to n(n-1)/2")

        if self.avgNumVertices < self.numVertexLabels:
            raise ValueError("Average number of vertices should be greater than or equal to number of vertex labels")

        if self.avgNumEdges < self.numEdgeLabels:
            raise ValueError("Average number of edges should be greater than or equal to number of edge labels")

    def generate(self):
        with open(self.outputFileName, 'w') as oFile:
            for i in range(self.numGraphs):
                numVertices = random.randint(max(self.avgNumVertices-3, 1), self.avgNumVertices+3)
                numEdges = random.randint(max(self.avgNumEdges-3, 0), self.avgNumEdges+3)

                if numVertices < numEdges:
                    numVertices = numEdges + 1

                if numEdges > numVertices * (numVertices - 1) / 2:
                    numEdges = numVertices * (numVertices - 1) // 2

                if numVertices < self.numVertexLabels:
                    numVertices = self.numVertexLabels

                if numEdges < self.numEdgeLabels:
                    numEdges = self.numEdgeLabels

                graph = {'nodes': [], 'edges': []}

                # Add vertices
                for j in range(numVertices):
                    graph['nodes'].append((j, random.randint(0, self.numVertexLabels-1)))

                # Ensure connectivity by creating a spanning tree first
                connectedNodes = set()
                connectedNodes.add(0)
                while len(connectedNodes) < numVertices:
                    u = random.choice(list(connectedNodes))
                    v = random.choice([node for node in range(numVertices) if node not in connectedNodes])
                    graph['edges'].append((u, v, random.randint(0, self.numEdgeLabels-1)))
                    connectedNodes.add(v)

                # Add remaining edges randomly
                additionalEdges = numEdges - (numVertices - 1)
                for _ in range(additionalEdges):
                    while True:
                        u = random.randint(0, numVertices-1)
                        v = random.randint(0, numVertices-1)
                        if u != v and (u, v, _) not in graph['edges'] and (v, u, _) not in graph['edges']:
                            graph['edges'].append((u, v, random.randint(0, self.numEdgeLabels-1)))
                            break

                if self.format == 'new':
                    self._writeGraphToFileNewFormat(graph, oFile)
                else:
                    self._writeGraphToFile(graph, oFile, i)


    def _writeGraphToFile(self, graph, oFile, i):
        oFile.write(f't # {i}\n')
        for node in graph['nodes']:
            oFile.write(f"v {node[0]} {node[1]}\n")
        for edge in graph['edges']:
            oFile.write(f"e {edge[0]} {edge[1]} {edge[2]}\n")

    def _writeGraphToFileNewFormat(self, graph, oFile):
        node_str = ' '.join(f"{node} {label}" for node, label in sorted(graph['nodes']))
        edge_str = ' '.join(f"{u} {v} {label}" for u, v, label in graph['edges'])
        oFile.write(f"{node_str} : {edge_str}\n")

if __name__ == "__main__":
    obj = SyntheticGraphGenerator(10, 2, 1, 1, 1, 'synthetic_graphs.txt', 'old')
    vis = graphs.graphDatabase('synthetic_graphs.txt')
    vis.plot()



# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from cudaAlgorithms import gPFMinerBit
#
#             obj = gPFMinerBit.gPFMinerBit("data.txt", 2, 3)
#
#             obj.run()
#
#             print(obj.getPatterns())
#
#             print(obj.getRuntime())
#
#             print(obj.getMemoryRSS())
#
#             print(obj.getMemoryUSS())
#
#             print(obj.getGPUMemory())
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""


import os
import sys
import csv
import time
import psutil
import numpy as np
import pycuda.autoinit
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from PAMI.periodicFrequentPattern.basic import abstract as _ab
import pandas as pd
from deprecated import deprecated


supportAndPeriod = SourceModule(r"""

__global__ void supportAndPeriod(unsigned long long int *bitArray, // containing transactions
                                unsigned long long int *support, // for support
                                unsigned long long int *period, // for period
                                unsigned long long int *thingsToCompare, // for things to compare
                                unsigned long long int *thingsToCompareIndex, // for things to compare index
                                unsigned long long int numberOfThingsToCompare, // for number of things to compare
                                unsigned long long int numberOfBits, // for number of bits
                                unsigned long long int numberOfElements, // for number of elements
                                unsigned long long int maxPeriod, // for max period
                                unsigned long long int maxTimeStamp){

        unsigned long long int threadIDX = blockIdx.x * blockDim.x + threadIdx.x;

        if (threadIDX > numberOfThingsToCompare-2) return;

        unsigned long long int holder = 0;
        unsigned long long int supportCounter = 0;
        unsigned long long int periodCounter = 0;
        unsigned long long int numbersCounter = 0;
        short int bitRepresentation[64];
        short int index = numberOfBits - 1;

        for(int i = 0; i < numberOfElements; i++){
            // intersection
            holder = bitArray[thingsToCompare[thingsToCompareIndex[threadIDX]] * numberOfElements + i];
            for (int j = thingsToCompareIndex[threadIDX]+1; j < thingsToCompareIndex[threadIDX + 1]; j++){
                holder = holder & bitArray[thingsToCompare[j] * numberOfElements + i];
            }

            // empty bitRepresentation
            for (int j = 0; j < 64; j++){
                bitRepresentation[j] = 0;
            }

            // conversion to bit representation
            index = numberOfBits - 1;
            while (holder > 0){
                bitRepresentation[index] = holder % 2;
                holder = holder / 2;
                index--;
            }

            // counting period
            for (int j = 0; j < numberOfBits; j++){
                periodCounter++;
                numbersCounter++;
                if (periodCounter > maxPeriod){
                    period[threadIDX] = periodCounter;
                    support[threadIDX] = supportCounter;
                    return;
                }
                if (bitRepresentation[j] == 1){
                    supportCounter++;
                    if (periodCounter > period[threadIDX]) period[threadIDX] = periodCounter;
                    periodCounter = 0;
                }
                if (numbersCounter == maxTimeStamp){
                    support[threadIDX] = supportCounter;
                    period[threadIDX] = periodCounter;
                    return;
                }
            }

        }
        support[threadIDX] = supportCounter;
        period[threadIDX] = periodCounter;
        return;

    }

"""
                                )


class gPFMinerBit:

    """
    :Description:   ECLAT is one of the fundamental algorithm to discover frequent patterns in a transactional database.
                    This algorithm applies ECLAT as well as calculates periodicity to find patterns in a temporal database.
                    This program employs downward closure property to  reduce the search space effectively.
                    This algorithm employs depth-first search technique to find the complete set of frequent patterns in a
                    temporal database.

    :Reference:   Penugonda Ravikumar 1,2,† , Palla Likhitha 2,† , Bathala Venus Vikranth Raj 2,† , Rage Uday Kiran 1,3,* ,† , Yutaka Watanobe 1 and Koji Zettsu 3. "Efficient Discovery of Periodic-Frequent Patterns in Columnar Temporal Databases.https://www.mdpi.com/2079-9292/10/12/1478

    :param  iFile: str :
                   Name of the Input file to mine complete set of periodic frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of periodic frequent pattern's
    :param  minSup: str:
                   Controls the minimum number of transactions in which every item must appear in a database.
    :param  maxPer: str:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

        filePath : str
             path of the file

        minSup : int
             minimum support

        maxPeriod : (int)
             maximum period

        sep : str, optional
         separator

        Patterns : dict, optional
            dictionary of the patterns. Defaults to {}.

        maxTimeStamp : int, optional
           maximum timestamp. Defaults to 0.

        __time : int, optional
           time taken to execute the algorithm. Defaults to 0.

        __memRSS : int, optional
           memory used by the program. Defaults to 0.

        __memUSS : int, optional
           memory used by the program. Defaults to 0.

        __GPU_MEM : int, optional
           GPU memory used by the program. Defaults to 0.

        __baseGPUMem : int, optional
           base GPU memory used by the program. Defaults to 0.


    **Importing this algorithm into a python program**
    ------------------------------------------------------------
    .. code-block:: python

         from cudaAlgorithms import gPFMinerBit

         obj = gPFMinerBit.gPFMinerBit("data.txt", 2, 3)

         obj.run()

         print(obj.getPatterns())

         print(obj.getRuntime())

         print(obj.getMemoryRSS())

         print(obj.getMemoryUSS())

         print(obj.getGPUMemory())

    **Running from the command line:**
    ---------------------------------------
   .. code-block:: console


       Format:

       (.venv) $ python3 gPFMinerBit.py data.txt 2 3 output.txt


               .. note:: minSup will be considered in percentage of database transactions

    **Credits:**
    ---------------
        This program is created by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.
        
    """

    # supportAndPeriod = supportAndPeriod.get_function("supportAndPeriod")

    def __init__(self, filePath, minSup, maxPeriod, sep="\t"):
        self.filePath = filePath
        self.sep = sep
        self.minSup = minSup
        self.Patterns = {}
        self.maxPeriod = maxPeriod
        self.__time = 0
        self.__memRSS = 0
        self.__memUSS = 0
        self.__GPU_MEM = 0
        self.__baseGPUMem = 0
        """
         :Methods:
        
        __readFile(): Read the file and return the data in a dictionary

        __getMaxPeriod(): Get the maximum period of the patterns

        __generateBitArray(): Generate the bit array

        getRuntime(): Get the runtime of the algorithm

        getMemoryRSS(): Get the memory used by the program

        getMemoryUSS(): Get the memory used by the program

        getGPUMemory(): Get the GPU memory used by the program

        getPatterns(): Get the patterns"""

    def __readFile(self):
        """
        Read the file and return the data in a dictionary

        :return: dict: dictionary of the data
        """
        basePattern = {}
        with open(self.filePath, "r") as f:
            reader = csv.reader(f, delimiter=self.sep)
            for row in reader:
                line = [str(item) for item in row if item != ""]
                self.maxTimeStamp = int(line[0])
                line = line[1:]
                for item in line:
                    if item not in basePattern:
                        basePattern[item] = [self.maxTimeStamp]
                    else:
                        basePattern[item].append(self.maxTimeStamp)

        basePattern = dict(
            sorted(basePattern.items(), key=lambda x: len(x[1]), reverse=True))

        return basePattern

    def __getMaxPeriod(self, array):
        """
        Get the maximum period of the array

        :param array (list): list of the array
        :return: int: maximum period
        """

        cur = 0
        per = 0
        # sort the array
        array = sorted(array)
        array.append(self.maxTimeStamp)

        for tid in array:
            per = max(per, tid - cur)
            if per > self.maxPeriod:
                break
            cur = tid
        return per

    def __generateBitArray(self, fileData):
        """
        Convert the dictionary into a bit array with valid candidates and return it with
        index which can be used to locate the candidates when multiplied with self.lengthOfArray

        :param fileData (dict): dictionary of the data

        :return:
            list: bit array
            list: index of the bit array
        """
        self.bitsToGen = 0
        self.numberOfBits = 64
        if self.maxTimeStamp % self.numberOfBits == 0:
            self.bitsToGen = self.maxTimeStamp // self.numberOfBits
        else:
            self.bitsToGen = self.maxTimeStamp // self.numberOfBits + 1
        index = 0
        index2id = []
        bitValues = []
        for key, value in fileData.items():
            if (len(value) >= self.minSup and self.__getMaxPeriod(value) <= self.maxPeriod):
                index2id.append(key)
                bits = [0] * self.bitsToGen * self.numberOfBits
                for item in value:
                    bits[item - 1] = 1
                bitVal = []
                prev = 0
                current = self.numberOfBits
                while current <= self.bitsToGen * self.numberOfBits:
                    # get bit array from prev to current
                    bit = bits[prev:current]
                    prev = current
                    current += self.numberOfBits
                    # convert the bit array to integer
                    bitVal.append(int("".join(str(x) for x in bit), 2))
                self.lengthOfArray = len(bitVal)
                bitValues.append(bitVal)
                index += 1
                self.Patterns[tuple([key])] = [len(
                    value), self.__getMaxPeriod(value)]
        # print(bitValues[0][0:10])
        bitValues = np.array(bitValues, dtype=np.uint64)
        # print(bitValues[0:10])
        gpuBitArray = cuda.mem_alloc(bitValues.nbytes)
        self.bvnb = bitValues.nbytes
        cuda.memcpy_htod(gpuBitArray, bitValues)
        return gpuBitArray, index2id

    @deprecated("It is recommended to use mine() instead of mine() for mining process")
    def startMine(self):
        """
        Start the mining process
        """
        startTime = time.time()
        data = self.__readFile()
        bitValues, index2id = self.__generateBitArray(data)

        keys = [[i] for i in range(len(index2id))]

        if len(keys) > 1:
            self.__eclat(bitValues, keys, index2id)

        print(
            "Periodic-Frequent patterns were generated successfully using gPFMinerBit"
        )
        self.__time = time.time() - startTime
        self.__memRSS = psutil.Process(os.getpid()).memory_info().rss
        self.__memUSS = psutil.Process(os.getpid()).memory_full_info().uss

    def Mine(self):
        """
        Start the mining process
        """
        startTime = time.time()
        data = self.__readFile()
        bitValues, index2id = self.__generateBitArray(data)

        keys = [[i] for i in range(len(index2id))]

        if len(keys) > 1:
            self.__eclat(bitValues, keys, index2id)

        print(
            "Periodic-Frequent patterns were generated successfully using gPFMinerBit"
        )
        self.__time = time.time() - startTime
        self.__memRSS = psutil.Process(os.getpid()).memory_info().rss
        self.__memUSS = psutil.Process(os.getpid()).memory_full_info().uss

    def __eclat(self, bitValues, keys, index2id):
        """
        Recursive Eclat

        :param bitValues (list): bit array
        :param keys (list): list of keys
        :param index2id (list): list of index to id
        """
        print("Number of Keys: " + str(len(keys)))
        locations = [0]
        newKeys = []
        for i in range(len(keys)):
            # print("Key: " + str(i))
            # iKey = keys[i]
            for j in range(i+1, len(keys)):
                # jKey = keys[j]
                if keys[i][:-1] == keys[j][:-1] and keys[i][-1] != keys[j][-1]:
                    newCan = keys[i] + [keys[j][-1]]
                    newKeys.append(newCan)
                    locations.append(locations[-1]+len(newKeys[-1]))
                else:
                    break

        locations = np.array(locations, dtype=np.uint64)
        newKeys = np.array(newKeys, dtype=np.uint64)
        newKeys = newKeys.flatten()
        support = np.zeros(len(newKeys), dtype=np.uint64)
        period = np.zeros(len(newKeys), dtype=np.uint64)

        if len(locations) > 1:
            totalMemory = support.nbytes + period.nbytes + \
                newKeys.nbytes + locations.nbytes + self.bvnb
            if totalMemory > self.__GPU_MEM:
                self.__GPU_MEM = totalMemory - self.__baseGPUMem

            gpuSupport = cuda.mem_alloc(support.nbytes)
            gpuPeriod = cuda.mem_alloc(period.nbytes)
            gpuNewKeys = cuda.mem_alloc(newKeys.nbytes)
            gpuLocations = cuda.mem_alloc(locations.nbytes)

            cuda.memcpy_htod(gpuSupport, support)
            cuda.memcpy_htod(gpuPeriod, period)
            cuda.memcpy_htod(gpuNewKeys, newKeys)
            cuda.memcpy_htod(gpuLocations, locations)

            # print("Number of New Keys: " + str(len(newKeys)))

            self.supportAndPeriod(bitValues, gpuSupport, gpuPeriod,
                                  gpuNewKeys, gpuLocations, np.uint64(
                                      len(locations)),
                                  np.uint64(self.numberOfBits), np.uint64(
                                      self.lengthOfArray),
                                  np.uint64(self.maxPeriod), np.uint64(
                                      self.maxTimeStamp),
                                  block=(32, 1, 1), grid=(len(locations)//32+1, 1, 1))

            cuda.memcpy_dtoh(support, gpuSupport)
            cuda.memcpy_dtoh(period, gpuPeriod)

            # free
            gpuSupport.free()
            gpuPeriod.free()
            gpuNewKeys.free()
            gpuLocations.free()

            keys = newKeys
            newKeys = []

            counter = 0
            top = 5
            for i in range(len(locations)-1):
                # print(support[i], period[i])
                # if i == 0:
                #     key = keys[locations[i]:locations[i+1]]
                #     nkey = sorted([index2id[key[i]] for i in range(len(key))])
                #     print(key, nkey, support[i], period[i])
                if support[i] >= self.minSup and period[i] <= self.maxPeriod:
                    key = keys[locations[i]:locations[i+1]]
                    
                    nkey = list([index2id[key[i]] for i in range(len(key))])
                    if tuple(nkey) not in self.Patterns:
                        self.Patterns[tuple(nkey)] = [support[i], period[i]]
                        newKeys.append(list(key))
                        # if counter < top:
                        #     print(nkey, support[i], period[i])
                        #     counter += 1


            keys = newKeys
            if len(keys) > 0:
                self.__eclat(bitValues, keys, index2id)

    def getRuntime(self):
        return self.__time

    def getMemoryRSS(self):
        return self.__memRSS

    def getMemoryUSS(self):
        return self.__memUSS

    def getGPUMemory(self):
        return self.__GPU_MEM

    def getPatterns(self):
        return self.Patterns
    
    def savePatterns(self, fileName):
        with open(fileName, "w") as f:
            for key in self.Patterns:
                f.write(str(key) + "\t" + str(self.Patterns[key][0]) + "\t" + str(self.Patterns[key][1]) + "\n")


if __name__ == "__main__":
    # filePath = "temporal_T10I4D100K.csv"
    # sep = "\t"
    # support = 50
    # maxPeriod = 25000
    # obj = gPFMinerBit(filePath, support, maxPeriod, sep)
    # obj.mine()
    # print("Time: ", obj.getRuntime())
    # print("Patterns: ", len(obj.getPatterns()))
    # print("GPU MEM: ", obj.getGPUMemory())
    # print("Mem RSS: ", obj.getMemoryRSS())
    # print("Mem USS: ", obj.getMemoryUSS())
    if len(sys.argv) != 6:
        print("Usage: python3 gPFMinerBit.py <input file> <min support> <max period> <separator> <output file>")
        sys.exit(1)
    filePath = sys.argv[1]
    support = int(sys.argv[2])
    maxPeriod = int(sys.argv[3])
    sep = sys.argv[4]
    output = sys.argv[5]
    obj = gPFMinerBit(filePath, support, maxPeriod, sep)
    obj.mine()
    obj.savePatterns(output)
    print("Time: ", obj.getRuntime())
    print("Patterns: ", len(obj.getPatterns()))
    print("GPU MEM: ", obj.getGPUMemory())
    print("Mem RSS: ", obj.getMemoryRSS())
    print("Mem USS: ", obj.getMemoryUSS())

